from abc import ABCMeta, abstractmethod
from typing import List

from ovos_bus_client import Message
from ovos_bus_client.message import dig_for_message
from ovos_utils import classproperty
from ovos_utils.fakebus import FakeBus
from ovos_utils.log import LOG
from ovos_utils.ocp import PlaybackType, TrackState, PlayerState, MediaState
from ovos_utils.process_utils import RuntimeRequirements


class AudioBackend(metaclass=ABCMeta):
    """Base class for all audio backend implementations.

    Arguments:
        config (dict): configuration dict for the instance
        bus (MessageBusClient): OpenVoiceOS messagebus emitter
    """

    def __init__(self, config=None, bus=None, name=None):
        self.name = name or self.__class__.__name__
        self._now_playing = None  # single uri
        self._tracks = []  # list of dicts for OCP entries
        self._idx = 0
        self._track_start_callback = None
        self.supports_mime_hints = False
        self.config = config or {}
        self.bus = bus or FakeBus()

    @classproperty
    def runtime_requirements(cls):
        """ skill developers should override this if they do not require connectivity
         some examples:
         IOT plugin that controls devices via LAN could return:
            scans_on_init = True
            RuntimeRequirements(internet_before_load=False,
                                 network_before_load=scans_on_init,
                                 requires_internet=False,
                                 requires_network=True,
                                 no_internet_fallback=True,
                                 no_network_fallback=False)
         online search plugin with a local cache:
            has_cache = False
            RuntimeRequirements(internet_before_load=not has_cache,
                                 network_before_load=not has_cache,
                                 requires_internet=True,
                                 requires_network=True,
                                 no_internet_fallback=True,
                                 no_network_fallback=True)
         a fully offline plugin:
            RuntimeRequirements(internet_before_load=False,
                                 network_before_load=False,
                                 requires_internet=False,
                                 requires_network=False,
                                 no_internet_fallback=True,
                                 no_network_fallback=True)
        """
        return RuntimeRequirements(internet_before_load=False,
                                   network_before_load=False,
                                   requires_internet=False,
                                   requires_network=False,
                                   no_internet_fallback=True,
                                   no_network_fallback=True)

    @property
    def playback_time(self) -> int:
        """ in milliseconds """
        return 0

    @abstractmethod
    def supported_uris(self) -> List[str]:
        """List of supported uri types.

        Returns:
            list: Supported uri's
        """

    @abstractmethod
    def play(self, repeat=False):
        """Start playback.

        Starts playing the first track in the playlist and will contiune
        until all tracks have been played.

        Arguments:
            repeat (bool): Repeat playlist, defaults to False
        """

    @abstractmethod
    def lower_volume(self):
        """Lower volume.

        This method is used to implement audio ducking. It will be called when
        OpenVoiceOS is listening or speaking to make sure the media playing isn't
        interfering.
        """

    @abstractmethod
    def restore_volume(self):
        """Restore normal volume.

        Called when to restore the playback volume to previous level after
        OpenVoiceOS has lowered it using lower_volume().
        """

    @abstractmethod
    def get_track_length(self) -> int:
        """
        getting the duration of the audio in miliseconds
        """

    @abstractmethod
    def get_track_position(self) -> int:
        """
        get current position in miliseconds
        """

    @abstractmethod
    def set_track_position(self, milliseconds):
        """Go to X position.
        Arguments:
           milliseconds (int): position to go to in milliseconds
        """

    @abstractmethod
    def pause(self):
        """Pause playback.

        Stops playback but may be resumed at the exact position the pause
        occured.
        """

    @abstractmethod
    def resume(self):
        """Resume paused playback.

        Resumes playback after being paused.
        """

    @abstractmethod
    def stop(self):
        """Stop playback.

        Stops the current playback.

        Returns:
            bool: True if playback was stopped, otherwise False
        """

    #####################
    # internals and default implementations
    def track_info(self) -> dict:
        """Request information of current playing track.
        Returns:
            Dict with track info.
        """
        return self._uri2meta(self._now_playing)

    def clear_list(self):
        """Clear playlist."""
        self._tracks = []
        self._idx = 0

    def add_list(self, tracks):
        """Add tracks to backend's playlist.

        Arguments:
            tracks (list): list of tracks.
        """
        tracks = tracks or []
        if isinstance(tracks, str):
            tracks = [tracks]
        elif not isinstance(tracks, list):
            raise ValueError
        if tracks and not self._tracks:
            self.load_track(tracks[0])
            self._idx = 0
        self._tracks += tracks

    def next(self):
        """Skip to next track in playlist."""
        self._idx += 1
        if self._idx < len(self._tracks):
            self.load_track(self._tracks[self._idx])
            self.play()
        else:
            LOG.error("no more tracks!")

    def previous(self):
        """Skip to previous track in playlist."""
        self._idx = max(self._idx - 1, 0)
        if self._idx < len(self._tracks):
            self.load_track(self._tracks[self._idx])
            self.play()
        else:
            LOG.error("already in first track!")

    def seek_forward(self, seconds=1):
        """Skip X seconds.

        Arguments:
            seconds (int): number of seconds to seek, if negative rewind
        """
        miliseconds = seconds * 1000
        new_pos = self.get_track_position() + miliseconds
        self.set_track_position(new_pos)

    def seek_backward(self, seconds=1):
        """Rewind X seconds.

        Arguments:
            seconds (int): number of seconds to seek, if negative jump forward.
        """
        miliseconds = seconds * 1000
        new_pos = self.get_track_position() - miliseconds
        self.set_track_position(new_pos)

    def set_track_start_callback(self, callback_func):
        """Register callback on track start.

        This method should be called as each track in a playlist is started.
        """
        self._track_start_callback = callback_func

    def shutdown(self):
        """Perform clean shutdown.

        Implements any audio backend specific shutdown procedures.
        """
        self.stop()

    def _format_msg(self, msg_type, msg_data=None):
        # this method ensures all skills are .forward from the utterance
        # that triggered the skill, this ensures proper routing and metadata
        msg_data = msg_data or {}
        msg = dig_for_message()
        if msg:
            msg = msg.forward(msg_type, msg_data)
        else:
            msg = Message(msg_type, msg_data)
        # at this stage source == skills, lets indicate audio service took over
        sauce = msg.context.get("source")
        if sauce == "skills":
            msg.context["source"] = "audio_service"
        return msg

    ############################
    # OCP extensions - new methods to improve compat with OCP
    @staticmethod
    def _uri2meta(uri):
        if isinstance(uri, list):
            uri = uri[0]
        try:
            from ovos_ocp_files_plugin.plugin import OCPFilesMetadataExtractor
            return OCPFilesMetadataExtractor.extract_metadata(uri)
        except:
            meta = {"uri": uri,
                    "skill_id": "mycroft.audio_interface",
                    "playback": PlaybackType.AUDIO,  # TODO mime type check
                    "status": TrackState.QUEUED_AUDIO,
                    }
        return meta

    def load_track(self, uri):
        """ This method is only used by ovos-core
        In ovos audio backends are single-track, playlists are handled by OCP
        """
        self._now_playing = uri
        LOG.debug(f"queuing for {self.__class__.__name__} playback: {uri}")
        self.bus.emit(Message("ovos.common_play.media.state",
                              {"state": MediaState.LOADING_MEDIA}))
        self.bus.emit(Message("ovos.common_play.track.state", {
            "state": TrackState.QUEUED_AUDIOSERVICE
        }))

    def ocp_sync_playback(self, playback_time):
        self.bus.emit(Message("ovos.common_play.playback_time",
                              {"position": playback_time,
                               "length": self.get_track_length()}))

    def ocp_start(self):
        """Emit OCP status events for play"""
        self.bus.emit(Message("ovos.common_play.player.state",
                              {"state": PlayerState.PLAYING}))
        self.bus.emit(Message("ovos.common_play.media.state",
                              {"state": MediaState.LOADED_MEDIA}))
        self.bus.emit(Message("ovos.common_play.track.state",
                              {"state": TrackState.PLAYING_AUDIOSERVICE}))

    def ocp_error(self):
        """Emit OCP status events for playback error"""
        if self._now_playing:
            self.bus.emit(Message("ovos.common_play.media.state",
                                  {"state": MediaState.INVALID_MEDIA}))
            self._now_playing = None

    def ocp_stop(self):
        """Emit OCP status events for stop"""
        if self._now_playing:
            self._now_playing = None
            self.bus.emit(Message("ovos.common_play.player.state",
                                  {"state": PlayerState.STOPPED}))
            self.bus.emit(Message("ovos.common_play.media.state",
                                  {"state": MediaState.END_OF_MEDIA}))

    def ocp_pause(self):
        """Emit OCP status events for pause"""
        if self._now_playing:
            self.bus.emit(Message("ovos.common_play.player.state",
                                  {"state": PlayerState.PAUSED}))

    def ocp_resume(self):
        """Emit OCP status events for resume"""
        if self._now_playing:
            self.bus.emit(Message("ovos.common_play.player.state",
                                  {"state": PlayerState.PLAYING}))
            self.bus.emit(Message("ovos.common_play.track.state",
                                  {"state": TrackState.PLAYING_AUDIOSERVICE}))


class RemoteAudioBackend(AudioBackend):
    """Base class for remote audio backends.

    RemoteAudioBackends will always be checked after the normal
    AudioBackends to make playback start locally by default.

    An example of a RemoteAudioBackend would be things like Chromecasts,
    mopidy servers, etc.
    """



import abc
from typing import List, Optional, Tuple, Dict, Union, Iterable

# Typing helpers for readability
try:
    import numpy as np
    EmbeddingsArray = np.ndarray
except ImportError:
    EmbeddingsArray = Iterable[Union[int, float]]
EmbeddingsTuple = Union[Tuple[str, float], Tuple[str, float, Dict]]


class EmbeddingsDB:
    """Base class for an embeddings database that supports storage, retrieval, and querying of embeddings."""

    @abc.abstractmethod
    def add_embeddings(self, key: str, embedding: EmbeddingsArray,
                       metadata: Optional[Dict[str, any]] = None) -> EmbeddingsArray:
        """Store 'embedding' under 'key' with associated metadata.

        Args:
            key (str): The unique key for the embedding.
            embedding (np.ndarray): The embedding vector to store.
            metadata (Optional[Dict[str, any]]): Optional metadata associated with the embedding.

        Returns:
            np.ndarray: The stored embedding.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_embeddings(self, key: str) -> EmbeddingsArray:
        """Retrieve embeddings stored under 'key'.

        Args:
            key (str): The unique key for the embedding.

        Returns:
            np.ndarray: The retrieved embedding.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def delete_embeddings(self, key: str) -> EmbeddingsArray:
        """Delete embeddings stored under 'key'.

        Args:
            key (str): The unique key for the embedding.

        Returns:
            np.ndarray: The deleted embedding.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def query(self, embeddings: EmbeddingsArray, top_k: int = 5,
              return_metadata: bool = False) -> List[EmbeddingsTuple]:
        """Return the top_k embeddings closest to the given 'embeddings'.

        Args:
            embeddings (np.ndarray): The embedding vector to query.
            top_k (int, optional): The number of top results to return. Defaults to 5.
            return_metadata (bool, optional): Whether to include metadata in the results. Defaults to False.

        Returns:
            List[EmbeddingsTuple]: List of tuples containing the key and distance, and optionally metadata.
        """
        raise NotImplementedError

    def distance(self, embeddings_a: EmbeddingsArray, embeddings_b: EmbeddingsArray, metric: str = "cosine",
                 alpha: float = 0.5,  # for alpha_divergence and tversky metrics
                 beta: float = 0.5,  # for tversky metric
                 p: float = 3,  # for minkowski and weighted_minkowski metrics
                 euclidean_weights: Optional[EmbeddingsArray] = None,  # required for weighted_euclidean and weighted_minkowski metrics
                 covariance_matrix: Optional[EmbeddingsArray] = None  # required for mahalanobis distance with user-defined covariance
                 ) -> float:
        """
        Calculate the distance between two embeddings vectors using the specified distance metric.

        Args:
            embeddings_a (np.ndarray): The first embedding vector.
            embeddings_b (np.ndarray): The second embedding vector.
            metric (str, optional): The distance metric to use. Defaults to "cosine".
                Supported metrics include:
                - "cosine": Cosine distance, 1 - cosine similarity. Useful for text similarity and high-dimensional data.
                - "euclidean": Euclidean distance, L2 norm of the difference. Commonly used in clustering and geometric distance.
                - "manhattan": Manhattan distance, L1 norm of the difference. Suitable for grid-based maps and robotics.
                - "chebyshev": Chebyshev distance, maximum absolute difference. Used for chessboard distance and pathfinding.
                - "minkowski": Minkowski distance, generalization of Euclidean and Manhattan distances. Parameterized by p, flexible use case.
                - "weighted_minkowski": Weighted Minkowski distance, a generalization of Minkowski with weights. Parameterized by `p`, uses `euclidean_weights`.
                - "hamming": Hamming distance, proportion of differing elements. Ideal for error detection and binary data.
                - "jaccard": Jaccard distance, 1 - Jaccard similarity (intersection over union). Used for set similarity and binary attributes.
                - "canberra": Canberra distance, weighted version of Manhattan distance. Sensitive to small changes, used in environmental data.
                - "braycurtis": Bray-Curtis distance, dissimilarity between non-negative vectors. Common in ecology and species abundance studies.
                - "mahalanobis": Mahalanobis distance, considering correlations (requires covariance matrix). Useful for multivariate outlier detection.
                - "pearson_correlation": Pearson correlation distance, 1 - Pearson correlation coefficient. Used in time series analysis and signal processing.
                - "spearman_rank": Spearman rank correlation distance, 1 - Spearman rank correlation coefficient. Measures rank correlation for non-linear monotonic relationships.
                - "wasserstein": Earth Mover's Distance (Wasserstein distance). Compares probability distributions or histograms.
                - "cosine_squared": Cosine squared distance, 1 - cosine similarity squared. For squared similarity in high-dimensional data.
                - "kl_divergence": Kullback-Leibler divergence, asymmetric measure of difference between distributions. Applied in information theory and probability distributions.
                - "bhattacharyya": Bhattacharyya distance, measure of overlap between statistical samples. Useful in classification and image processing.
                - "hellinger": Hellinger distance, measure of similarity between two probability distributions. Applied in statistical inference.
                - "ruzicka": Ruzicka distance, similarity measure for non-negative vectors. Used in ecology and species abundance.
                - "kulczynski": Kulczynski distance, used in ecology to compare similarity. Suitable for ecological studies and species distribution.
                - "sorensen": Sørensen distance, another name for Dice distance. Applied in binary data comparison and text similarity.
                - "chi_squared": Chi-squared distance, used for comparing categorical data distributions. Suitable for categorical data analysis and distribution comparison.
                - "jensen_shannon": Jensen-Shannon divergence, symmetrized and smoothed version of KL divergence. Used in information theory and probability distributions.
                - "squared_euclidean": Squared Euclidean distance, square of the Euclidean distance. Useful for clustering algorithms and geometric distance.
                - "weighted_euclidean": Weighted Euclidean distance, L2 norm with weights. Applied when features have different scales or importance.
                - "log_cosh": Log-Cosh distance, log of the hyperbolic cosine of the difference. Robust to outliers.
                - "tanimoto": Tanimoto coefficient, similarity measure for binary vectors. Used for binary data comparison.
                - "rao": Rao's Quadratic Entropy, measure of divergence between distributions. Useful for comparing probability distributions.
                - "gower": Gower distance, handles mixed types of data. Applied in cases with numerical and categorical data.
                - "tversky": Tversky index, generalization of Jaccard and Dice for asymmetrical comparison. Parameterized by alpha and beta.
                - "alpha_divergence": Alpha divergence, generalized divergence measure. Parameterized by alpha, used for comparing distributions.
                - "kendall_tau": Kendall's Tau distance: 1 - Kendall Tau correlation coefficient. Use case: Rank correlation for ordinal data
                - "renyi_divergence": Generalized divergence measure. Use case: Comparing probability distributions
                - "total_variation":  Measure of divergence between distributions. Use case: Probability distributions, statistical inference

            alpha (float, optional): Parameter for `tversky` and `alpha_divergence` metrics. Default is 0.5.
            beta (float, optional): Parameter for `tversky` metric. Default is 0.5.
            p (float, optional): Parameter for `minkowski` and `weighted_minkowski` metrics. Default is 3.
            euclidean_weights (Optional[np.ndarray], optional): Weights for `weighted_euclidean` and `weighted_minkowski` metrics. Must be provided if using these metrics. Default is None.
            covariance_matrix (Optional[np.ndarray], optional): Covariance matrix for `mahalanobis` distance. Must be provided if using this metric. Default is None.

        Returns:
            float: The calculated distance between the two embedding vectors.

        Raises:
            ValueError: If the specified metric is unsupported or requires parameters not provided.
        """
        if metric == "cosine":
            # Cosine distance: 1 - cosine similarity
            # Use case: Text similarity, high-dimensional data
            dot = np.dot(embeddings_a, embeddings_b)
            norma = np.linalg.norm(embeddings_a)
            normb = np.linalg.norm(embeddings_b)
            cos = dot / (norma * normb)
            return 1 - cos
        elif metric == "euclidean":
            # Euclidean distance: L2 norm of the difference
            # Use case: Geometric distance, clustering
            return np.linalg.norm(embeddings_a - embeddings_b)
        elif metric == "manhattan":
            # Manhattan distance: L1 norm of the difference
            # Use case: Grid-based maps, robotics
            return np.sum(np.abs(embeddings_a - embeddings_b))
        elif metric == "chebyshev":
            # Chebyshev distance: Maximum absolute difference
            # Use case: Chessboard distance, pathfinding
            return np.max(np.abs(embeddings_a - embeddings_b))
        elif metric == "minkowski":
            # Minkowski distance: Generalization of Euclidean and Manhattan distances
            # Use case: Flexible distance metric, parameterized by p
            return np.sum(np.abs(embeddings_a - embeddings_b) ** p) ** (1 / p)
        elif metric == "weighted_minkowski":
            # Weighted Minkowski distance: Generalization of Minkowski distance with weights
            # Use case: Flexible distance metric with weighted dimensions
            if euclidean_weights is None:
                raise ValueError("euclidean_weights must be provided for weighted_minkowski metric")
            return np.sum(euclidean_weights * np.abs(embeddings_a - embeddings_b) ** p) ** (1 / p)
        elif metric == "hamming":
            # Hamming distance: Proportion of differing elements
            # Use case: Error detection, binary data
            return np.mean(embeddings_a != embeddings_b)
        elif metric == "jaccard":
            # Jaccard distance: 1 - Jaccard similarity (intersection over union)
            # Use case: Set similarity, binary attributes
            intersection = np.sum(np.minimum(embeddings_a, embeddings_b))
            union = np.sum(np.maximum(embeddings_a, embeddings_b))
            return 1 - intersection / union
        elif metric == "canberra":
            # Canberra distance: Weighted version of Manhattan distance
            # Use case: Environmental data, sensitive to small changes
            return np.sum(np.abs(embeddings_a - embeddings_b) / (np.abs(embeddings_a) + np.abs(embeddings_b)))
        elif metric == "braycurtis":
            # Bray-Curtis distance: Dissimilarity between non-negative vectors
            # Use case: Ecology, species abundance
            return np.sum(np.abs(embeddings_a - embeddings_b)) / np.sum(np.abs(embeddings_a + embeddings_b))
        elif metric == "mahalanobis":
            # Mahalanobis distance: Distance considering correlations (requires covariance matrix)
            # Use case: Multivariate outlier detection
            if covariance_matrix is None:
                covariance_matrix = np.cov(embeddings_a, embeddings_b, rowvar=False)
            inv_cov_matrix = np.linalg.inv(covariance_matrix)
            delta = embeddings_a - embeddings_b
            return np.sqrt(np.dot(np.dot(delta.T, inv_cov_matrix), delta))
        elif metric == "pearson_correlation":
            # Correlation distance: 1 - Pearson correlation coefficient
            # Use case: Time series analysis, signal processing
            mean_a = np.mean(embeddings_a)
            mean_b = np.mean(embeddings_b)
            centered_a = embeddings_a - mean_a
            centered_b = embeddings_b - mean_b
            norm_a = np.linalg.norm(centered_a)
            norm_b = np.linalg.norm(centered_b)
            correlation = np.dot(centered_a, centered_b) / (norm_a * norm_b)
            return 1 - correlation
        elif metric == "spearman_rank":
            # Spearman rank correlation distance: 1 - Spearman rank correlation coefficient.
            # Use case: Measures the rank correlation between two vectors. Useful for non-linear monotonic relationships.
            rank_a = np.argsort(np.argsort(embeddings_a))
            rank_b = np.argsort(np.argsort(embeddings_b))
            return 1 - np.corrcoef(rank_a, rank_b)[0, 1]
        elif metric == "wasserstein":
            # Earth Mover's Distance (Wasserstein distance)
            # Use case: Comparing probability distributions or histograms
            arr1_sorted = np.sort(embeddings_a)
            arr2_sorted = np.sort(embeddings_b)
            cdf1 = np.cumsum(arr1_sorted) / np.sum(arr1_sorted)
            cdf2 = np.cumsum(arr2_sorted) / np.sum(arr2_sorted)
            return np.sum(np.abs(cdf1 - cdf2))
        elif metric == "cosine_squared":
            # Cosine squared distance: 1 - cosine similarity squared
            # Use case: Squared similarity, high-dimensional data
            dot = np.dot(embeddings_a, embeddings_b)
            norma = np.linalg.norm(embeddings_a)
            normb = np.linalg.norm(embeddings_b)
            cos = dot / (norma * normb)
            return 1 - cos ** 2
        elif metric == "kl_divergence":
            # Kullback-Leibler divergence: Asymmetric measure of difference between distributions
            # Use case: Information theory, probability distributions
            return np.sum(embeddings_a * np.log(embeddings_a / embeddings_b))
        elif metric == "bhattacharyya":
            # Bhattacharyya distance: Measure of overlap between statistical samples
            # Use case: Classification, image processing
            bc = np.sum(np.sqrt(embeddings_a * embeddings_b))
            return -np.log(bc)
        elif metric == "hellinger":
            # Hellinger distance: Measure of similarity between two probability distributions
            # Use case: Probability distributions, statistical inference
            return np.sqrt(0.5 * np.sum((np.sqrt(embeddings_a) - np.sqrt(embeddings_b)) ** 2))
        elif metric == "ruzicka":
            # Ruzicka distance: Similarity measure for non-negative vectors
            # Use case: Ecology, species abundance
            return 1 - np.sum(np.minimum(embeddings_a, embeddings_b)) / np.sum(np.maximum(embeddings_a, embeddings_b))
        elif metric == "kulczynski":
            # Kulczynski distance: Measure used in ecology to compare similarity
            # Use case: Ecological studies, species distribution
            return np.sum(np.abs(embeddings_a - embeddings_b)) / np.sum(np.minimum(embeddings_a, embeddings_b))
        elif metric == "sorensen":
            # Sørensen distance: Another name for Dice distance
            # Use case: Binary data comparison, text similarity
            intersection = np.sum(embeddings_a * embeddings_b)
            return 1 - (2 * intersection) / (np.sum(embeddings_a) + np.sum(embeddings_b))
        elif metric == "chi_squared":
            # Chi-squared distance: Used for comparing categorical data distributions
            # Use case: Categorical data analysis, distribution comparison
            return np.sum((embeddings_a - embeddings_b) ** 2 / (embeddings_a + embeddings_b))
        elif metric == "jensen_shannon":
            # Jensen-Shannon divergence: Symmetrized and smoothed version of KL divergence
            # Use case: Information theory, probability distributions
            m = 0.5 * (embeddings_a + embeddings_b)
            return 0.5 * (np.sum(embeddings_a * np.log(embeddings_a / m)) + np.sum(embeddings_b * np.log(embeddings_b / m)))
        elif metric == "squared_euclidean":
            # Squared Euclidean distance: Square of the Euclidean distance
            # Use case: Clustering algorithms, geometric distance
            return np.sum((embeddings_a - embeddings_b) ** 2)
        elif metric == "weighted_euclidean":
            # Weighted Euclidean distance: L2 norm with weights
            # Use case: Features with different scales or importance
            if euclidean_weights is None:
                raise ValueError("euclidean_weights must be provided for weighted_euclidean metric")
            return np.sqrt(np.sum(euclidean_weights * (embeddings_a - embeddings_b) ** 2))
        elif metric == "log_cosh":
            # Log-Cosh distance: Log of the hyperbolic cosine of the difference
            # Use case: Robustness to outliers
            return np.sum(np.log(np.cosh(embeddings_a - embeddings_b)))
        elif metric == "tanimoto":
            # Tanimoto coefficient: Similarity measure for binary vectors
            # Use case: Binary data comparison
            intersection = np.sum(embeddings_a * embeddings_b)
            return 1 - intersection / (np.sum(embeddings_a) + np.sum(embeddings_b) - intersection)
        elif metric == "rao":
            # Rao's Quadratic Entropy: Measure of divergence between distributions
            # Use case: Comparing probability distributions
            p = embeddings_a / np.sum(embeddings_a)
            q = embeddings_b / np.sum(embeddings_b)
            return np.sum((p - q) ** 2 / (p + q))
        elif metric == "gower":
            # Gower distance: Handles mixed types of data
            # Use case: Mixed data types (numerical and categorical)
            numerical_part = np.sum(np.abs(embeddings_a - embeddings_b)) / len(embeddings_a)
            categorical_part = np.mean(embeddings_a != embeddings_b)
            return numerical_part + categorical_part
        elif metric == "tversky":
            # Tversky index: Generalization of Jaccard and Dice for asymmetrical comparison
            intersection = np.sum(np.minimum(embeddings_a, embeddings_b))
            return 1 - intersection / (intersection + alpha * np.sum(embeddings_a - embeddings_b) + beta * np.sum(embeddings_b - embeddings_a))
        elif metric == "alpha_divergence":
            # Alpha divergence: Generalized divergence measure
            p = embeddings_a / np.sum(embeddings_a)
            q = embeddings_b / np.sum(embeddings_b)
            return np.sum((p ** alpha - q ** alpha) / (alpha * (p + q) ** alpha))
        elif metric == "kendall_tau":
            # Kendall's Tau distance: 1 - Kendall Tau correlation coefficient
            # Use case: Rank correlation for ordinal data
            concordant = np.sum((embeddings_a > embeddings_b) == (embeddings_b > embeddings_a))
            discordant = np.sum((embeddings_a > embeddings_b) != (embeddings_b > embeddings_a))
            return 1 - (concordant - discordant) / (concordant + discordant)
        elif metric == "renyi_divergence":
            # Renyi Divergence: Generalized divergence measure
            # Use case: Comparing probability distributions
            p = embeddings_a / np.sum(embeddings_a)
            q = embeddings_b / np.sum(embeddings_b)
            return 1 / (1 - alpha) * np.log(np.sum((p ** alpha + q ** alpha) / 2))
        elif metric == "total_variation":
            # Total Variation distance: Measure of divergence between distributions
            # Use case: Probability distributions, statistical inference
            p = embeddings_a / np.sum(embeddings_a)
            q = embeddings_b / np.sum(embeddings_b)
            return 0.5 * np.sum(np.abs(p - q))
        else:
            raise ValueError("Unsupported metric")


class TextEmbeddingsStore:
    """A store for text embeddings interfacing with the embeddings database."""

    def __init__(self, db: EmbeddingsDB):
        """Initialize the text embeddings store.

        Args:
            db (EmbeddingsDB): The embeddings database instance.
        """
        self.db = db

    @abc.abstractmethod
    def get_text_embeddings(self, text: str) -> EmbeddingsArray:
        """Convert text to its corresponding embeddings.

        Args:
            text (str): The input text to be converted.

        Returns:
            np.ndarray: The resulting embeddings.
        """
        raise NotImplementedError

    def add_document(self, document: str, metadata: Optional[Dict[str, any]] = None) -> None:
        """Add a document and its embeddings to the database.

        Args:
            document (str): The document to add.
            metadata (Optional[Dict[str, any]]): Optional metadata associated with the document.
        """
        embeddings = self.get_text_embeddings(document)
        self.db.add_embeddings(document, embeddings, metadata)

    def delete_document(self, document: str) -> None:
        """Delete a document and its embeddings from the database.

        Args:
            document (str): The document to delete.
        """
        self.db.delete_embeddings(document)

    def query(self, document: str, top_k: int = 5,
              return_metadata: bool = False) -> List[Tuple[str, float]]:
        """Query the database for the top_k closest embeddings to the document.

        Args:
            document (str): The document to query.
            top_k (int, optional): The number of top results to return. Defaults to 5.
            return_metadata (bool, optional): Whether to include metadata in the results. Defaults to False.

        Returns:
            List[Tuple[str, float]]: List of tuples containing the document and distance.
        """
        embeddings = self.get_text_embeddings(document)
        return self.db.query(embeddings, top_k,
                             return_metadata=return_metadata)

    def distance(self, text_a: str, text_b: str, metric: str = "cosine") -> float:
        """Calculate the distance between embeddings of two texts.

        Args:
            text_a (str): The first text.
            text_b (str): The second text.
            metric (str, optional): The distance metric to use. Defaults to "cosine".

        Returns:
            float: The calculated distance.
        """
        emb_a = self.get_text_embeddings(text_a)
        emb_b = self.get_text_embeddings(text_b)
        return self.db.distance(emb_a, emb_b, metric)


class FaceEmbeddingsStore:
    """A store for face embeddings interfacing with the embeddings database."""

    def __init__(self, db: EmbeddingsDB):
        """Initialize the face embeddings store.

        Args:
            db (EmbeddingsDB): The embeddings database instance.
        """
        self.db = db

    @abc.abstractmethod
    def get_face_embeddings(self, frame: EmbeddingsArray) -> EmbeddingsArray:
        """Convert an image frame to its corresponding face embeddings.

        Args:
            frame (np.ndarray): The input image frame containing a face.

        Returns:
            np.ndarray: The resulting face embeddings.
        """
        raise NotImplementedError

    def add_face(self, user_id: str, frame: EmbeddingsArray, metadata: Optional[Dict[str, any]] = None) -> EmbeddingsArray:
        """Add a face and its embeddings to the database.

        Args:
            user_id (str): The unique user ID.
            frame (np.ndarray): The image frame containing the face.
            metadata (Optional[Dict[str, any]]): Optional metadata associated with the face.

        Returns:
            np.ndarray: The stored face embeddings.
        """
        embeddings = self.get_face_embeddings(frame)
        return self.db.add_embeddings(user_id, embeddings, metadata)

    def delete_face(self, user_id: str) -> EmbeddingsArray:
        """Delete a face and its embeddings from the database.

        Args:
            user_id (str): The unique user ID.

        Returns:
            np.ndarray: The deleted face embeddings.
        """
        return self.db.delete_embeddings(user_id)

    def predict(self, frame: EmbeddingsArray, top_k: int = 3, thresh: float = 0.15) -> Optional[str]:
        """Return the top predicted face closest to the given frame.

        Args:
            frame (np.ndarray): The input image frame containing a face.
            top_k (int, optional): The number of top results to return. Defaults to 3.
            thresh (float, optional): The threshold for prediction. Defaults to 0.15.

        Returns:
            Optional[str]: The predicted user ID or None if the best match exceeds the threshold.
        """
        matches = self.query(frame, top_k)
        if not matches:
            return None
        best_match = min(matches, key=lambda k: k[1])
        if best_match[1] > thresh:
            return None
        return best_match[0]

    def query(self, frame: EmbeddingsArray, top_k: int = 5,
                return_metadata: bool = False) -> List[Tuple[str, float]]:
        """Query the database for the top_k closest face embeddings to the frame.

        Args:
            frame (np.ndarray): The input image frame containing a face.
            top_k (int, optional): The number of top results to return. Defaults to 5.
            return_metadata (bool, optional): Whether to include metadata in the results. Defaults to False.

        Returns:
            List[Tuple[str, float]]: List of tuples containing the user ID and distance.
        """
        embeddings = self.get_face_embeddings(frame)
        return self.db.query(embeddings, top_k,
                             return_metadata=return_metadata)

    def distance(self, face_a: EmbeddingsArray, face_b: EmbeddingsArray, metric: str = "cosine") -> float:
        """Calculate the distance between embeddings of two faces.

        Args:
            face_a (np.ndarray): The first face embedding.
            face_b (np.ndarray): The second face embedding.
            metric (str, optional): The distance metric to use. Defaults to "cosine".

        Returns:
            float: The calculated distance.
        """
        emb_a = self.get_face_embeddings(face_a)
        emb_b = self.get_face_embeddings(face_b)
        return self.db.distance(emb_a, emb_b, metric)


class VoiceEmbeddingsStore:
    """A store for voice embeddings interfacing with the embeddings database."""

    def __init__(self, db: EmbeddingsDB):
        """Initialize the voice embeddings store.

        Args:
            db (EmbeddingsDB): The embeddings database instance.
        """
        self.db = db

    @staticmethod
    def audiochunk2array(audio_bytes: bytes) -> EmbeddingsArray:
        """Convert audio buffer to a normalized float32 NumPy array.

        Args:
            audio_bytes (bytes): The audio data buffer.

        Returns:
            np.ndarray: The normalized float32 audio array.
        """
        audio_as_np_int16 = np.frombuffer(audio_bytes, dtype=np.int16)
        audio_as_np_float32 = audio_as_np_int16.astype(np.float32)
        # Normalise float32 array so that values are between -1.0 and +1.0
        max_int16 = 2 ** 15
        return audio_as_np_float32 / max_int16

    @abc.abstractmethod
    def get_voice_embeddings(self, audio_data: EmbeddingsArray) -> EmbeddingsArray:
        """Convert audio data to its corresponding voice embeddings.

        Args:
            audio_data (np.ndarray): The input audio data.

        Returns:
            np.ndarray: The resulting voice embeddings.
        """
        raise NotImplementedError

    def add_voice(self, user_id: str, audio_data: EmbeddingsArray, metadata: Optional[Dict[str, any]] = None) -> EmbeddingsArray:
        """Add a voice and its embeddings to the database.

        Args:
            user_id (str): The unique user ID.
            audio_data (np.ndarray): The input audio data.
            metadata (Optional[Dict[str, any]]): Optional metadata associated with the voice.

        Returns:
            np.ndarray: The stored voice embeddings.
        """
        embeddings = self.get_voice_embeddings(audio_data)
        return self.db.add_embeddings(user_id, embeddings, metadata)

    def delete_voice(self, user_id: str) -> EmbeddingsArray:
        """Delete a voice and its embeddings from the database.

        Args:
            user_id (str): The unique user ID.

        Returns:
            np.ndarray: The deleted voice embeddings.
        """
        return self.db.delete_embeddings(user_id)

    def predict(self, audio_data: EmbeddingsArray, top_k: int = 3, thresh: float = 0.75) -> Optional[str]:
        """Return the top predicted voice closest to the given audio_data.

        Args:
            audio_data (np.ndarray): The input audio data.
            top_k (int, optional): The number of top results to return. Defaults to 3.
            thresh (float, optional): The threshold for prediction. Defaults to 0.75.

        Returns:
            Optional[str]: The predicted user ID or None if the best match exceeds the threshold.
        """
        matches = self.query(audio_data, top_k)
        if not matches:
            return None
        best_match = min(matches, key=lambda k: k[1])
        if best_match[1] > thresh:
            return None
        return best_match[0]

    def query(self, audio_data: EmbeddingsArray, top_k: int = 5,
                return_metadata: bool = False) -> List[Tuple[str, float]]:
        """Query the database for the top_k closest voice embeddings to the audio_data.

        Args:
            audio_data (np.ndarray): The input audio data.
            top_k (int, optional): The number of top results to return. Defaults to 5.
            return_metadata (bool, optional): Whether to include metadata in the results. Defaults to False.

        Returns:
            List[Tuple[str, float]]: List of tuples containing the user ID and distance.
        """
        embeddings = self.get_voice_embeddings(audio_data)
        return self.db.query(embeddings, top_k,
                             return_metadata=return_metadata)

    def distance(self, voice_a: EmbeddingsArray, voice_b: EmbeddingsArray, metric: str = "cosine") -> float:
        """Calculate the distance between embeddings of two voices.

        Args:
            voice_a (np.ndarray): The first voice embedding.
            voice_b (np.ndarray): The second voice embedding.
            metric (str, optional): The distance metric to use. Defaults to "cosine".

        Returns:
            float: The calculated distance.
        """
        emb_a = self.get_voice_embeddings(voice_a)
        emb_b = self.get_voice_embeddings(voice_b)
        return self.db.distance(emb_a, emb_b, metric)



import json
import os
from hashlib import md5
from typing import Optional

from ovos_utils.log import LOG
from ovos_utils.xdg_utils import xdg_data_home

from ovos_plugin_manager.templates.hotwords import HotWordEngine
from ovos_plugin_manager.utils import PluginTypes, PluginConfigTypes


def find_wake_word_plugins() -> dict:
    """
    Find all installed plugins
    @return: dict plugin names to entrypoints
    """
    from ovos_plugin_manager.utils import find_plugins
    return find_plugins(PluginTypes.WAKEWORD)


def load_wake_word_plugin(module_name: str) -> type(HotWordEngine):
    """
    Get an uninstantiated class for the requested module_name
    @param module_name: Plugin entrypoint name to load
    @return: Uninstantiated class
    """
    from ovos_plugin_manager.utils import load_plugin
    return load_plugin(module_name, PluginTypes.WAKEWORD)


def get_ww_configs() -> dict:
    """
    Get valid plugin configurations by plugin name
    @return: dict plugin names to list of dict configurations
    """
    from ovos_plugin_manager.utils.config import load_configs_for_plugin_type
    return load_configs_for_plugin_type(PluginTypes.WAKEWORD)


def get_ww_module_configs(module_name: str) -> dict:
    """
    Get valid configurations for the specified plugin
    @param module_name: plugin to get configuration for
    @return: dict configurations by ww name (if provided)
    """
    # WW plugins return {ww_name: [list of config dicts]}
    from ovos_plugin_manager.utils.config import load_plugin_configs
    return load_plugin_configs(module_name, PluginConfigTypes.WAKEWORD)


def get_ww_lang_configs(lang: str, include_dialects: bool = False) -> dict:
    """
    Get a dict of plugin names to list valid configurations for the requested
    lang.
    @param lang: Language to get configurations for
    @param include_dialects: consider configurations in different locales
    @return: dict {`plugin_name`: `valid_configs`]}
    """
    from ovos_plugin_manager.utils.config import get_plugin_language_configs
    return get_plugin_language_configs(PluginTypes.WAKEWORD, lang,
                                       include_dialects)


def get_ww_supported_langs() -> dict:
    """
    Return a dict of plugin names to list supported languages
    @return: dict plugin names to list supported languages
    """
    from ovos_plugin_manager.utils.config import get_plugin_supported_languages
    return get_plugin_supported_languages(PluginTypes.WAKEWORD)


def get_hotwords_config(config: dict = None) -> dict:
    """
    Get relevant configuration for factory methods
    @param config: global Configuration OR plugin class-specific configuration
    @return: plugin class-specific configuration
    """
    from ovos_plugin_manager.utils.config import get_plugin_config
    return get_plugin_config(config, "hotwords")


def get_ww_id(plugin_name, ww_name, ww_config):
    ww_hash = md5(json.dumps(ww_config, sort_keys=True).encode("utf-8")).hexdigest()
    return f"{plugin_name}_{ww_name}_{ww_hash}"


def scan_wws():
    ww_ids = {}
    raise NotImplementedError("plugin wake word metadata reporting is WIP")
    return ww_ids


def get_wws(scan=False):
    if scan:
        scan_wws()
    ww_ids = {}
    for lang in get_ww_supported_langs():
        WW_FOLDER = f"{xdg_data_home()}/OPM/ww_configs/{lang}"
        for voice in os.listdir(WW_FOLDER):
            with open(f"{WW_FOLDER}/{voice}") as f:
                ww_ids[voice] = json.load(f)
    return ww_ids


class OVOSWakeWordFactory:

    @staticmethod
    def get_class(hotword: str, config: Optional[dict] = None) -> type:
        """
        Get the plugin class for the specified hotword
        @param hotword: string hotword to load
        @param config: optional global configuration
        @return: Uninitialized hotword class
        """
        hotword_config = get_hotwords_config(config)
        if hotword not in hotword_config:
            LOG.warning(f"{hotword} not in {hotword_config}! "
                        f"Returning base HotWordEngine")
            return HotWordEngine
        ww_module = hotword_config[hotword]["module"]
        return load_wake_word_plugin(ww_module)

    @staticmethod
    def load_module(module: str, hotword: str, hotword_config: dict) -> HotWordEngine:
        """
        Get an initialized HotWordEngine using the specified module and hotword
        @param module: hotword plugin to load (not parsed)
        @param hotword: string hotword to load
        @param hotword_config: configuration for the specified `hotword`.
            Equivalent to Configuration()['hotwords'][hotword]
        @param loop: Unused
        @return: Initialized HotWordEngine
        """
        # config here is config['hotwords'][module]
        LOG.info(f'Loading "{hotword}" wake word via {module} with '
                 f'config: {hotword_config}')
        config = {"hotwords": {hotword: hotword_config}}
        clazz = OVOSWakeWordFactory.get_class(hotword, config)
        if clazz is None:
            raise ImportError(f'Wake Word {hotword} with module {module} '
                              f'failed to load')
        LOG.info(f'Loaded the Wake Word {hotword} with module {module}')
        return clazz(hotword, hotword_config)

    @classmethod
    def create_hotword(cls, hotword: str = "hey mycroft",
                       config: Optional[dict] = None) -> HotWordEngine:
        """
        Get an initialized HotWordEngine by configured name
        @param hotword: string hotword to load
        @param config: optional global configuration
        @return: Initialized HotWordEngine
        """
        ww_configs = get_hotwords_config(config)
        if hotword not in ww_configs:
            LOG.warning(f"replace ` ` in {hotword} with `_`")
            hotword = hotword.replace(' ', '_')
        ww_config = ww_configs.get(hotword)
        module = ww_config.get("module", "pocketsphinx")
        try:
            return cls.load_module(module, hotword, ww_config)
        except Exception as e:
            LOG.error(f"Failed to load hotword: {hotword} - {module}")
            LOG.exception(e)
            fallback_ww = ww_config.get("fallback_ww")
            if fallback_ww in ww_configs and fallback_ww != hotword:
                LOG.info(f"Attempting to load fallback ww instead: {fallback_ww}")
                return cls.create_hotword(fallback_ww, config)
            raise



from abc import ABCMeta, abstractmethod

from ovos_bus_client.message import Message
from ovos_utils.log import LOG
from ovos_utils.fakebus import FakeBus
from ovos_utils.ocp import MediaState, PlayerState, TrackState


class MediaBackend(metaclass=ABCMeta):
    """Base class for all OCP media backend implementations.

    Media backends are single-track, playlists are handled by OCP

   Arguments:
       config (dict): configuration dict for the instance
       bus (MessageBusClient): Mycroft messagebus emitter
   """

    def __init__(self, config=None, bus=None):
        if MediaState is None:
            raise RuntimeError("Please update to ovos-utils~=0.1.")
        self._now_playing = None  # single uri
        self._track_start_callback = None
        self.supports_mime_hints = False
        self.config = config or {}
        self.bus = bus or FakeBus()
        self.meta = {}

    def set_track_start_callback(self, callback_func):
        """Register callback on track start.

        This method should be called as each track in a playlist is started.
        """
        self._track_start_callback = callback_func

    def load_track(self, uri: str, metadata: dict = None):
        self._now_playing = uri
        self.meta.update(metadata or {})
        LOG.debug(f"queuing for {self.__class__.__name__} playback: {uri}")
        self.bus.emit(Message("ovos.common_play.media.state",
                              {"state": MediaState.LOADED_MEDIA}))

    def ocp_start(self):
        """Emit OCP status events for play"""
        self.bus.emit(Message("ovos.common_play.player.state",
                              {"state": PlayerState.PLAYING}))
        self.bus.emit(Message("ovos.common_play.media.state",
                              {"state": MediaState.LOADED_MEDIA}))
        self.play()

    def ocp_error(self):
        """Emit OCP status events for playback error"""
        if self._now_playing:
            self._now_playing = None
            self.bus.emit(Message("ovos.common_play.media.state",
                                  {"state": MediaState.INVALID_MEDIA}))
            self.bus.emit(Message("ovos.common_play.player.state",
                                  {"state": PlayerState.STOPPED}))

    def ocp_stop(self):
        """Emit OCP status events for stop"""
        if self._now_playing:
            self._now_playing = None
            self.bus.emit(Message("ovos.common_play.player.state",
                                  {"state": PlayerState.STOPPED}))
            self.bus.emit(Message("ovos.common_play.media.state",
                                  {"state": MediaState.END_OF_MEDIA}))
            self.stop()

    def ocp_pause(self):
        """Emit OCP status events for pause"""
        if self._now_playing:
            self.bus.emit(Message("ovos.common_play.player.state",
                                  {"state": PlayerState.PAUSED}))
            self.pause()

    def ocp_resume(self):
        """Emit OCP status events for resume"""
        if self._now_playing:
            self.bus.emit(Message("ovos.common_play.player.state",
                                  {"state": PlayerState.PLAYING}))
            self.bus.emit(Message("ovos.common_play.track.state",
                                  {"state": TrackState.PLAYING_AUDIO}))
            self.resume()

    @property
    def playback_time(self):
        return 0

    @abstractmethod
    def supported_uris(self):
        """List of supported uri types.

        Returns:
            list: Supported uri's
        """

    @abstractmethod
    def play(self):
        """Start playback.

        Starts playing the first track in the playlist and will contiune
        until all tracks have been played.
        """

    @abstractmethod
    def stop(self):
        """Stop playback.

        Stops the current playback.

        Returns:
            bool: True if playback was stopped, otherwise False
        """

    @abstractmethod
    def pause(self):
        """Pause playback.

        Stops playback but may be resumed at the exact position the pause
        occured.
        """

    @abstractmethod
    def resume(self):
        """Resume paused playback.

        Resumes playback after being paused.
        """

    @abstractmethod
    def lower_volume(self):
        """Lower volume.

        This method is used to implement audio ducking. It will be called when
        OpenVoiceOS is listening or speaking to make sure the media playing isn't
        interfering.
        """

    @abstractmethod
    def restore_volume(self):
        """Restore normal volume.

        Called when to restore the playback volume to previous level after
        OpenVoiceOS has lowered it using lower_volume().
        """

    @abstractmethod
    def get_track_length(self) -> int:
        """
        getting the duration of the audio in milliseconds
        """

    @abstractmethod
    def get_track_position(self) -> int:
        """
        get current position in milliseconds
        """

    @abstractmethod
    def set_track_position(self, milliseconds):
        """
        go to position in milliseconds
          Args:
                milliseconds (int): number of milliseconds of final position
        """

    def seek_forward(self, seconds=1):
        """Skip X seconds.

        Arguments:
            seconds (int): number of seconds to seek, if negative rewind
        """
        miliseconds = seconds * 1000
        new_pos = self.get_track_position() + miliseconds
        self.set_track_position(new_pos)

    def seek_backward(self, seconds=1):
        """Rewind X seconds.

        Arguments:
            seconds (int): number of seconds to seek, if negative jump forward.
        """
        miliseconds = seconds * 1000
        new_pos = self.get_track_position() - miliseconds
        self.set_track_position(new_pos)

    def track_info(self):
        """Get info about current playing track.

        Returns:
            dict: Track info containing atleast the keys artist and album.
        """
        return self.meta

    def shutdown(self):
        """Perform clean shutdown.

        Implements any audio backend specific shutdown procedures.
        """
        self.stop()


class AudioPlayerBackend(MediaBackend):
    """ for audio"""

    def load_track(self, uri, metadata: dict = None):
        super().load_track(uri, metadata)
        self.bus.emit(Message("ovos.common_play.track.state",
                              {"state": TrackState.QUEUED_AUDIO}))

    def ocp_start(self):
        """Emit OCP status events for play"""
        super().ocp_start()
        self.bus.emit(Message("ovos.common_play.track.state",
                              {"state": TrackState.PLAYING_AUDIO}))


class RemoteAudioPlayerBackend(AudioPlayerBackend):
    """Base class for remote audio backends.

    RemoteAudioBackends will always be checked after the normal
    AudioBackends to make playback start locally by default.

    An example of a RemoteAudioBackend would be things like mopidy servers, etc.
    """


class VideoPlayerBackend(MediaBackend):
    """ for video"""
    def load_track(self, uri, metadata: dict = None):
        super().load_track(uri, metadata)
        self.bus.emit(Message("ovos.common_play.track.state",
                              {"state": TrackState.QUEUED_VIDEO}))

    def ocp_start(self):
        """Emit OCP status events for play"""
        super().ocp_start()
        self.bus.emit(Message("ovos.common_play.track.state",
                              {"state": TrackState.PLAYING_VIDEO}))


class RemoteVideoPlayerBackend(VideoPlayerBackend):
    """Base class for remote audio backends.

    RemoteVideoBackends will always be checked after the normal
    VideoBackends to make playback start locally by default.

    An example of a RemoteVideoBackend would be things like Chromecasts, etc.
    """


class WebPlayerBackend(MediaBackend):
    """ for web pages"""

    def load_track(self, uri, metadata: dict = None):
        super().load_track(uri, metadata)
        self.bus.emit(Message("ovos.common_play.track.state",
                              {"state": TrackState.QUEUED_WEBVIEW}))

    def ocp_start(self):
        """Emit OCP status events for play"""
        super().ocp_start()
        self.bus.emit(Message("ovos.common_play.track.state",
                              {"state": TrackState.PLAYING_WEBVIEW}))


class RemoteWebPlayerBackend(WebPlayerBackend):
    """Base class for remote web backends.

    RemoteWebBackends will always be checked after the normal
    VideoBackends to make playback start locally by default.

    An example of a RemoteWebBackend would be
    things that can render a webpage in a different machine
    """



import abc
from typing import Tuple, List
from ovos_bus_client.session import SessionManager
from ovos_utils import classproperty
from ovos_utils.lang import standardize_lang_tag
from ovos_utils.process_utils import RuntimeRequirements


Tag = Tuple[int, int, str, str]  # start_idx, end_idx, word, tag


class PosTagger:
    def __init__(self, config=None):
        self.config = config or {}

    @classproperty
    def runtime_requirements(cls):
        """ skill developers should override this if they do not require connectivity
         some examples:
         IOT plugin that controls devices via LAN could return:
            scans_on_init = True
            RuntimeRequirements(internet_before_load=False,
                                 network_before_load=scans_on_init,
                                 requires_internet=False,
                                 requires_network=True,
                                 no_internet_fallback=True,
                                 no_network_fallback=False)
         online search plugin with a local cache:
            has_cache = False
            RuntimeRequirements(internet_before_load=not has_cache,
                                 network_before_load=not has_cache,
                                 requires_internet=True,
                                 requires_network=True,
                                 no_internet_fallback=True,
                                 no_network_fallback=True)
         a fully offline plugin:
            RuntimeRequirements(internet_before_load=False,
                                 network_before_load=False,
                                 requires_internet=False,
                                 requires_network=False,
                                 no_internet_fallback=True,
                                 no_network_fallback=True)
        """
        return RuntimeRequirements(internet_before_load=False,
                                   network_before_load=False,
                                   requires_internet=False,
                                   requires_network=False,
                                   no_internet_fallback=True,
                                   no_network_fallback=True)

    @property
    def lang(self) -> str:
        lang = self.config.get("lang") or SessionManager.get().lang
        return standardize_lang_tag(lang)

    @abc.abstractmethod
    def postag(self, spans, lang=None) -> List[Tag]:
        raise NotImplementedError()



from abc import abstractmethod


class AbstractSwitches:
    @property
    @abstractmethod
    def capabilities(self) -> dict:
        """
        Return a dict of capabilities this object supports
        """

    @abstractmethod
    def on_action(self):
        """
        Override to do something when the `action` button is pressed.
        """
        pass

    @abstractmethod
    def on_vol_up(self):
        """
        Override to do something when the `volume up` button is pressed.
        """
        pass

    @abstractmethod
    def on_vol_down(self):
        """
        Override to do something when the `volume down` button is pressed.
        """
        pass

    @abstractmethod
    def on_mute(self):
        """
        Override to do something when `mute` switch is activated.
        """
        pass

    @abstractmethod
    def on_unmute(self):
        """
        Override to do something when `mute` switch is deactivated.
        """
        pass

    @abstractmethod
    def shutdown(self):
        """
        Perform any cleanup.
        """

    def get_capabilities(self) -> dict:
        """
        Backwards-compatible method to return `self.capabilities`
        """
        return self.capabilities



from ovos_plugin_manager.templates.transformers import IntentTransformer
from ovos_plugin_manager.utils import PluginTypes


def find_intent_transformer_plugins() -> dict:
    """
    Find all installed plugins
    @return: dict plugin names to entrypoints
    """
    from ovos_plugin_manager.utils import find_plugins
    return find_plugins(PluginTypes.INTENT_TRANSFORMER)


def load_intent_transformer_plugin(module_name: str) -> \
        type(IntentTransformer):
    """
    Get an uninstantiated class for the requested module_name
    @param module_name: Plugin entrypoint name to load
    @return: Uninstantiated class
    """
    from ovos_plugin_manager.utils import load_plugin
    return load_plugin(module_name, PluginTypes.INTENT_TRANSFORMER)



from ovos_plugin_manager.utils.config import get_plugin_config
from ovos_utils.log import LOG

from ovos_config import Configuration
from ovos_plugin_manager.templates.language import LanguageTranslator, \
    LanguageDetector
from ovos_plugin_manager.utils import PluginTypes, PluginConfigTypes


def find_tx_plugins() -> dict:
    """
    Find all installed plugins
    @return: dict plugin names to entrypoints
    """
    from ovos_plugin_manager.utils import find_plugins
    return find_plugins(PluginTypes.TRANSLATE)


def load_tx_plugin(module_name: str) -> type(LanguageTranslator):
    """
    Get an uninstantiated class for the requested module_name
    @param module_name: Plugin entrypoint name to load
    @return: Uninstantiated class
    """
    from ovos_plugin_manager.utils import load_plugin
    return load_plugin(module_name, PluginTypes.TRANSLATE)


def get_tx_configs() -> dict:
    """
    Get valid plugin configurations by plugin name
    @return: dict plugin names to list of dict configurations
    """
    from ovos_plugin_manager.utils.config import load_configs_for_plugin_type
    return load_configs_for_plugin_type(PluginTypes.TRANSLATE)


def get_tx_module_configs(module_name: str):
    """
    Get valid configurations for the specified plugin
    @param module_name: plugin to get configuration for
    @return: list of dict configurations (if provided)  TODO Validate type
    """
    from ovos_plugin_manager.utils.config import load_plugin_configs
    return load_plugin_configs(module_name, PluginConfigTypes.TRANSLATE)


def find_lang_detect_plugins():
    """
    Find all installed plugins
    @return: dict plugin names to entrypoints
    """
    from ovos_plugin_manager.utils import find_plugins
    return find_plugins(PluginTypes.LANG_DETECT)


def load_lang_detect_plugin(module_name: str) -> type(LanguageDetector):
    """
    Get an uninstantiated class for the requested module_name
    @param module_name: Plugin entrypoint name to load
    @return: Uninstantiated class
    """
    from ovos_plugin_manager.utils import load_plugin
    return load_plugin(module_name, PluginTypes.LANG_DETECT)


def get_lang_detect_configs() -> dict:
    """
    Get valid plugin configurations by plugin name
    @return: dict plugin names to list of dict configurations
    """
    from ovos_plugin_manager.utils.config import load_configs_for_plugin_type
    return load_configs_for_plugin_type(PluginTypes.LANG_DETECT)


def get_lang_detect_module_configs(module_name: str):
    """
    Get valid configurations for the specified plugin
    @param module_name: plugin to get configuration for
    @return: list of dict configurations (if provided)  TODO Validate type
    """
    from ovos_plugin_manager.utils.config import load_plugin_configs
    return load_plugin_configs(module_name, PluginConfigTypes.LANG_DETECT)


class OVOSLangDetectionFactory:

    @staticmethod
    def get_class(config=None):
        """
        Factory method to get a Language Detector class based on configuration.

        Configuration contains a `language` section with
        the name of a LangDetection module to be read by this method.

        "language": {
            "detection_module": <engine_name>
        }
        """
        config = config or Configuration()
        if "language" in config:
            config = config["language"]
        lang_module = config.get("detection_module", config.get("module"))
        if not lang_module:
            raise ValueError("`language.detection_module` not configured")
        return load_lang_detect_plugin(lang_module)

    @classmethod
    def create(cls, config=None) -> LanguageDetector:
        """
        Factory method to create a LangDetection engine based on configuration

        Configuration contains a `language` section with
        the name of a LangDetection module to be read by this method.

        "language": {
            "detection_module": <engine_name>
        }
        """
        config = config or Configuration()
        if "language" in config:
            config = config["language"]
        lang_module = config.get("detection_module", config.get("module"))
        cfg = config.get(lang_module, {})
        fallback = cfg.get("fallback_module")
        try:
            config["module"] = lang_module
            clazz = OVOSLangDetectionFactory.get_class(config)
            if clazz is None:
                raise ValueError(f"Failed to load module: {lang_module}")
            LOG.info(f'Loaded the Language Detection plugin {lang_module}')
            return clazz(config=get_plugin_config(config, "language",
                                                  lang_module))
        except Exception:
            LOG.exception(f'Language Detection plugin {lang_module} could not be loaded!')
            if fallback in config and fallback != lang_module:
                LOG.info(f"Attempting to load fallback plugin instead: {fallback}")
                config["detection_module"] = fallback
                return cls.create(config)
            raise


class OVOSLangTranslationFactory:

    @staticmethod
    def get_class(config=None):
        """
        Factory method to get a Language Translator class based on configuration.

        Configuration contains a `language` section with
        the name of a Translation module to be read by this method.

        "language": {
            "translation_module": <engine_name>
        }
        """
        config = config or Configuration()
        if "language" in config:
            config = config["language"]
        lang_module = config.get("translation_module", config.get("module"))
        if not lang_module:
            raise ValueError("`language.translation_module` not configured")
        return load_tx_plugin(lang_module)

    @classmethod
    def create(cls, config=None) -> LanguageTranslator:
        """
        Factory method to create a LangTranslation engine based on configuration

        Configuration contains a `language` section with
        the name of a Translation module to be read by this method.

        "language": {
            "translation_module": <engine_name>
        }
        """
        config = config or Configuration()
        if "language" in config:
            config = config["language"]
        lang_module = config.get("translation_module", config.get("module"))
        cfg = config.get(lang_module, {})
        fallback = cfg.get("fallback_module")
        try:
            config["module"] = lang_module
            clazz = OVOSLangTranslationFactory.get_class(config)
            if clazz is None:
                raise ValueError(f"Failed to load module: {lang_module}")
            LOG.info(f'Loaded the Language Translation plugin {lang_module}')
            return clazz(config=get_plugin_config(config, "language",
                                                  lang_module))
        except Exception:
            LOG.exception(f'Language Translation plugin {lang_module} could not be loaded!')
            if fallback in config and fallback != lang_module:
                LOG.info(f"Attempting to load fallback plugin instead: {fallback}")
                config["translation_module"] = fallback
                return cls.create(config)
            raise



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Common functions for loading plugins."""
import time
from collections import deque
from enum import Enum
from threading import Event, Lock
from typing import Optional
import warnings
import pkg_resources
from ovos_utils.log import LOG, log_deprecation, deprecated


class PluginTypes(str, Enum):
    TRIPLES = "opm.triples"
    PIPELINE = "opm.pipeline"
    EMBEDDINGS = "opm.embeddings"
    FACE_EMBEDDINGS = "opm.embeddings.face"
    VOICE_EMBEDDINGS = "opm.embeddings.voice"
    TEXT_EMBEDDINGS = "opm.embeddings.text"
    GUI = "ovos.plugin.gui"  # TODO rename "opm.gui"
    PHAL = "ovos.plugin.phal"  # TODO rename "opm.phal"
    ADMIN = "ovos.plugin.phal.admin"  # TODO rename "opm.phal.admin"
    SKILL = "ovos.plugin.skill"  # TODO rename "opm.skill"
    MIC = "ovos.plugin.microphone"  # TODO rename "opm.microphone"
    VAD = "ovos.plugin.VAD"  # TODO rename "opm.vad"
    PHONEME = "ovos.plugin.g2p"  # TODO rename "opm.g2p"
    AUDIO2IPA = "ovos.plugin.audio2ipa"  # TODO rename "opm.audio2ipa"
    AUDIO = 'mycroft.plugin.audioservice'  # DEPRECATED
    STT = 'mycroft.plugin.stt'  # TODO rename "opm.stt"
    TTS = 'mycroft.plugin.tts'  # TODO rename "opm.tts"
    WAKEWORD = 'mycroft.plugin.wake_word'  # TODO rename "opm.wake_word"
    TRANSLATE = "neon.plugin.lang.translate"  # TODO rename "opm.lang.translate"
    LANG_DETECT = "neon.plugin.lang.detect"  # TODO rename "opm.lang.detect"
    UTTERANCE_TRANSFORMER = "neon.plugin.text"  # TODO rename "opm.transformer.text"
    METADATA_TRANSFORMER = "neon.plugin.metadata"  # TODO rename "opm.transformer.metadata"
    AUDIO_TRANSFORMER = "neon.plugin.audio"  # TODO rename "opm.transformer.audio"
    DIALOG_TRANSFORMER = "opm.transformer.dialog"
    TTS_TRANSFORMER = "opm.transformer.tts"
    INTENT_TRANSFORMER = "opm.transformer.intent"
    QUESTION_SOLVER = "neon.plugin.solver"  # TODO rename "opm.solver.question"
    CHAT_SOLVER = "opm.solver.chat"
    TLDR_SOLVER = "opm.solver.summarization"
    ENTAILMENT_SOLVER = "opm.solver.entailment"
    MULTIPLE_CHOICE_SOLVER = "opm.solver.multiple_choice"
    READING_COMPREHENSION_SOLVER = "opm.solver.reading_comprehension"
    COREFERENCE_SOLVER = "intentbox.coreference"  # TODO rename "opm.coreference"
    KEYWORD_EXTRACTION = "intentbox.keywords"  # TODO rename "opm.keywords"
    UTTERANCE_SEGMENTATION = "intentbox.segmentation"  # TODO rename "opm.segmentation"
    TOKENIZATION = "intentbox.tokenization"  # TODO rename "opm.tokenization"
    POSTAG = "intentbox.postag"  # TODO rename "opm.postag"
    STREAM_EXTRACTOR = "ovos.ocp.extractor"  # TODO rename "opm.ocp.extractor"
    AUDIO_PLAYER = "opm.media.audio"
    VIDEO_PLAYER = "opm.media.video"
    WEB_PLAYER = "opm.media.web"
    PERSONA = "opm.plugin.persona"  # personas are a dict, they have no config because they ARE a config


class PluginConfigTypes(str, Enum):
    TRIPLES = "opm.triples.config"
    PIPELINE = "opm.pipeline.config"
    EMBEDDINGS = "opm.embeddings.config"
    FACE_EMBEDDINGS = "opm.embeddings.face.config"
    VOICE_EMBEDDINGS = "opm.embeddings.voice.config"
    TEXT_EMBEDDINGS = "opm.embeddings.text.config"
    GUI = "ovos.plugin.gui.config"
    PHAL = "ovos.plugin.phal.config"
    ADMIN = "ovos.plugin.phal.admin.config"
    SKILL = "ovos.plugin.skill.config"
    VAD = "ovos.plugin.VAD.config"
    MIC = "ovos.plugin.microphone.config"
    PHONEME = "ovos.plugin.g2p.config"
    AUDIO2IPA = "ovos.plugin.audio2ipa.config"
    AUDIO = 'mycroft.plugin.audioservice.config'
    STT = 'mycroft.plugin.stt.config'
    TTS = 'mycroft.plugin.tts.config'
    WAKEWORD = 'mycroft.plugin.wake_word.config'
    TRANSLATE = "neon.plugin.lang.translate.config"
    LANG_DETECT = "neon.plugin.lang.detect.config"
    UTTERANCE_TRANSFORMER = "neon.plugin.text.config"
    METADATA_TRANSFORMER = "neon.plugin.metadata.config"
    AUDIO_TRANSFORMER = "neon.plugin.audio.config"
    DIALOG_TRANSFORMER = "opm.transformer.dialog.config"
    TTS_TRANSFORMER = "opm.transformer.tts.config"
    INTENT_TRANSFORMER = "opm.transformer.intent.config"
    QUESTION_SOLVER = "neon.plugin.solver.config"
    CHAT_SOLVER = "opm.solver.chat.config"
    TLDR_SOLVER = "opm.solver.summarization.config"
    ENTAILMENT_SOLVER = "opm.solver.entailment.config"
    MULTIPLE_CHOICE_SOLVER = "opm.solver.multiple_choice.config"
    READING_COMPREHENSION_SOLVER = "opm.solver.reading_comprehension.config"
    COREFERENCE_SOLVER = "intentbox.coreference.config"
    KEYWORD_EXTRACTION = "intentbox.keywords.config"
    UTTERANCE_SEGMENTATION = "intentbox.segmentation.config"
    TOKENIZATION = "intentbox.tokenization.config"
    POSTAG = "intentbox.postag.config"
    STREAM_EXTRACTOR = "ovos.ocp.extractor.config"
    AUDIO_PLAYER = "opm.media.audio.config"
    VIDEO_PLAYER = "opm.media.video.config"
    WEB_PLAYER = "opm.media.web.config"


def find_plugins(plug_type: PluginTypes = None) -> dict:
    """
    Finds all plugins matching specific entrypoint type.

    Arguments:
        plug_type (str): plugin entrypoint string to retrieve

    Returns:
        dict mapping plugin names to plugin entrypoints
    """
    entrypoints = {}
    if not plug_type:
        plugs = list(PluginTypes)
    elif isinstance(plug_type, str):
        plugs = [plug_type]
    else:
        plugs = plug_type
    for plug in plugs:
        for entry_point in _iter_entrypoints(plug):
            try:
                entrypoints[entry_point.name] = entry_point.load()
                if entry_point.name not in entrypoints:
                    LOG.debug(f"Loaded plugin entry point {entry_point.name}")
            except Exception as e:
                if entry_point not in find_plugins._errored:
                    find_plugins._errored.append(entry_point)
                    # NOTE: this runs in a loop inside skills manager, this would endlessly spam logs
                    LOG.error(f"Failed to load plugin entry point {entry_point}: "
                              f"{e}")
    return entrypoints


find_plugins._errored = []


def _iter_entrypoints(plug_type: Optional[str]):
    """
    Return an iterator containing all entrypoints of the requested type
    @param plug_type: entrypoint name to load
    @return: iterator of all entrypoints
    """
    try:
        from importlib_metadata import entry_points
        for entry_point in entry_points(group=plug_type):
            yield entry_point
    except ImportError:
        for entry_point in pkg_resources.iter_entry_points(plug_type):
            yield entry_point


def load_plugin(plug_name: str, plug_type: Optional[PluginTypes] = None):
    """Load a specific plugin from a specific plugin type.

    Arguments:
        plug_type: (str) plugin type name. Ex. "mycroft.plugin.tts".
        plug_name: (str) specific plugin name (else consider all plugin types)

    Returns:
        Loaded plugin Object or None if no matching object was found.
    """
    plugins = find_plugins(plug_type)
    if plug_name in plugins:
        return plugins[plug_name]
    plug_type = plug_type or "all plugin types"
    LOG.warning(f'Could not find the plugin {plug_type}.{plug_name}')
    return None


@deprecated("normalize_lang has been deprecated! update to 'from ovos_utils.lang import standardize_lang_tag'", "1.0.0")
def normalize_lang(lang):
    warnings.warn(
        "update to 'from ovos_utils.lang import standardize_lang_tag'",
        DeprecationWarning,
        stacklevel=2,
    )
    from ovos_utils.lang import standardize_lang_tag
    return standardize_lang_tag(lang)


class ReadWriteStream:
    """
    Class used to support writing binary audio data at any pace,
    with an optional maximum buffer size
    """

    def __init__(self, s=b'', chop_samples=-1, max_size=None):
        if chop_samples != -1:
            log_deprecation("'chop_samples' kwarg has been deprecated and will be ignored", "1.0.0")
        self.buffer = deque(s)
        self.write_event = Event()
        self.lock = Lock()
        self.max_size = max_size  # Introduce max size

    def __len__(self):
        with self.lock:
            return len(self.buffer)

    def read(self, n=-1, timeout=None):
        with self.lock:
            if n == -1 or n > len(self.buffer):
                n = len(self.buffer)

        end_time = time.time() + timeout if timeout is not None else float('inf')

        while True:
            with self.lock:
                if len(self.buffer) >= n:
                    chunk = bytes([self.buffer.popleft() for _ in range(n)])
                    return chunk

            remaining_time = None
            if timeout is not None:
                remaining_time = end_time - time.time()
                if remaining_time <= 0:
                    return b''

            self.write_event.clear()
            self.write_event.wait(remaining_time)

    def write(self, s):
        with self.lock:
            self.buffer.extend(s)
            if self.max_size is not None:
                while len(self.buffer) > self.max_size:
                    self.buffer.popleft()  # Discard oldest data to maintain max size
        self.write_event.set()

    def flush(self):
        """Makes compatible with sys.stdout"""
        pass

    def clear(self):
        with self.lock:
            self.buffer.clear()
